	/* @(#)write.c	1.17  9/15/91 */

	/*-------------------------------------------------------------
	|							      |
	|	PROGRAM:  ntscable                                    |
	|							      |
	|	MODULE:   write.c                                     |
	|							      |
	|	MACHINE:  Sun 3/60                                    |
	|							      |
	|	STARTED:  20-APR-89        BY:  J.C. Wathey           |
	|							      |
	|	REVISED:  9/15/91         BY:  JCW                    |
	|	          5/26/92         BY:  Z. Mainen              |
	|							      |
	|	STATUS:      incomplete or untested		      |
	|                    compiles; partly tested		      |
	|                    runs; revisions in progress	      |
	|                 -> runs; stable version		      |
	|							      |
	|       CONTAINS: routines write_geometry() and               |
	|                 and write_neuron_description()              |
	|                                                             |
	|       COMPILE:                                              |
	|                                                             |
	|       (use makefile)                                        |
	|                                                             |
	-------------------------------------------------------------*/

/*--------------------------------------- ZFM Changes ---------------

 **** THE FOLLOWING CHANGES AFFECT *NEURON* SYNTAX OUTPUT ONLY! ****

-  added provision for axon
   -c flag creates axon
   changes affect here, ntscable.c, cmdline.c
-  neuron files as well as cable files can be generated with 
     accuracy of 0, 1, or 2! (see "xlate.c")
   changes affect cmdline.c, xlate.c
-  added provision for explicit hoc-language information about
     the cell in neuron_syntax output (see "write.c")
   -e flag writes explicit info
   changes affect here, cmdline.c, write.c
-  creates a soma with name defined in "ntscable.h" rather than
     hard-coded "soma"

---------------------------------------------------------------------*/


/*--------------------------------------- HEADER FILES --------------*/
#include "tf.h"
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include "ntscable.h"		       /* global defs & declarations */

static char sepline[] = 
"----------------------------------------------------------------";



/*-------------------------------------------------------------------*/
int write_geometry( fp, gobi_fp, gobi_ln_fp, neuron_ptr, cable_ptr, neurite_list )


	/* Writes to fp a "geometry" subroutine in hoc syntax, using
	the info in the *neuron_ptr and either *cable_ptr or 
	*neurite_list, depending on the syntax requested for the
	output file.  Returns TRUE if error; FALSE otherwise. */


    FILE      * fp,
              * gobi_fp,
              * gobi_ln_fp;
    NEURON    * neuron_ptr; 
    CABLE     * cable_ptr;
    NEURITE   * neurite_list;

{
				       /*----- functions called -----*/

				       /*----- extern variables -----*/

				       /*----- local  variables -----*/

				       /*----- start function -------*/

					 /* check translation syntax */
    switch (neuron_ptr->destination) {

	case O_MULTI:
	case O_CABLE:
            return(write_cable_geometry( fp, 
					 neuron_ptr, 
					 cable_ptr));
	    break;                    /*This break wasn't here before */

	default:
	case O_NEURON:
            if (neuron_ptr->cell_name == NULL)
	      return(write_neuron_geometry( fp, gobi_fp, gobi_ln_fp,
					   neuron_ptr, 
					   neurite_list));
	    else
	      return(write_multi_neuron_geometry(fp, gobi_fp, gobi_ln_fp,
						 neuron_ptr,
						 neurite_list));
    }

}

/*-------------------------------------------------------------------*/

static int write_neuron_geometry( fp, gobi_fp, gobi_ln_fp,
			          neuron_ptr, 
			          neurite_list )


	/* Writes to fp a "geometry" subroutine in hoc syntax, using
	the info in the *neuron_ptr and *neurite_list.  Returns TRUE 
	if error; FALSE otherwise. */


    FILE      * fp,
              * gobi_fp,
              * gobi_ln_fp;
    NEURON    * neuron_ptr; 
    NEURITE   * neurite_list;


{
				       /*----- functions called -----*/
				       /*----- extern variables -----*/
				       /*----- local  variables -----*/

    static char for_loop_i[] = "    for i = %d,%d {\n",
                for_loop_j[] = "for j = 1, fscan() {\n",
                read_3D_points[] =
		"pt3dadd(fscan(),fscan(),fscan(),fscan())";

    int			neurite_i,
                        section_i, 
                        xyzd_i;

    NEURITE	      * neurite_ptr,
		      * neurite_list_end;
    SECTION	      * section_ptr,
		      * section_list_end;
    XYZD	      * xyzd_ptr,
		      * xyzd_list_end;
    char     soma_name[100], axon_name[100], dend_name[100], temp_buf[100];

    printf ("Insinde write_neuron_geometry().\n");

				       /*----- start function -------*/

    neurite_list_end = neurite_list + neuron_ptr->num_primary_neurites;

                                       /* handle cell names if present */

    if (neuron_ptr->cell_name != NULL) {
      sprintf(soma_name,"%s_%s",neuron_ptr->cell_name, SOMA_NAME);
      sprintf(axon_name,"%s_%s",neuron_ptr->cell_name, AXON_NAME);
      for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {
                 /*I had to create a temporary buffer for the new ->name, this 
                   was due to the fact that you can not sprintf to a string
                   that you need to use as an argument to sprintf.  so the
                   former line:
                    sprintf (neurite_ptr->name "%s_%s", neuron_ptr->cell_name, 
                             neurite_ptr->name) was crashing.         Ritter */
         sprintf(temp_buf, "%s_%s",neuron_ptr->cell_name, neurite_ptr->name);
         strcpy (neurite_ptr->name, temp_buf);
       }
    } else {
      sprintf(soma_name,"%s",SOMA_NAME);
      sprintf(axon_name,"%s",AXON_NAME);
    }

    write_neuron_comment(fp, neuron_ptr);

    if (neuron_ptr->write_explicit_info) 
      write_neuron_explicit_info(fp, neuron_ptr, neurite_list);
    

    						  /* create sections */

    fprintf(fp,"\t\t\t\t\t/* create sections */\n");
    fprintf(fp,"create    %s",soma_name);

                                                   /* create an axon */

    if (neuron_ptr->create_axon) {
      fprintf(fp,",\\\n          %s",axon_name);
    }


    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

	fprintf(fp, ",\\\n          %s[%d]",
		    neurite_ptr->name,
		    neurite_ptr->num_sections);
    }
    fprintf(fp,"\n\n");

					  /* main geometry procedure */

    fprintf(fp, "\n\n/*%s*/\n", sepline);
    fprintf(fp,"proc geometry() { local i, j\n\n");

						    /* soma geometry */

    fprintf(fp,"\t\t\t\t\t\t/* soma geometry */\n");

    fprintf(fp, "    %s {\n",soma_name);
    fprintf(fp, "        nseg = 1\n");
    fprintf(fp, "        pt3dclear()\n");
    fprintf(fp, "        %s", for_loop_j );
    fprintf(fp, "            %s\n", read_3D_points);
    fprintf(fp, "        }\n");
    fprintf(fp,"    }\n\n");

                                            /* minimal axon geometry */
    if (neuron_ptr->create_axon) {
      fprintf(fp,"\t\t\t\t\t\t/* axon geometry */\n");
      fprintf(fp, "    %s {\n",axon_name);
      fprintf(fp, "        nseg = 1\n");
      fprintf(fp, "        pt3dclear()\n");
      fprintf(fp, "        pt3dadd(0,0,0,1e-4)\n");
      fprintf(fp, "        pt3dadd(0,0,1e-4,1e-4)\n");
      fprintf(fp, "    }\n");
      fprintf(fp,"\t\t\t\t\t\t/* connect axon */\n");
      fprintf(fp, "    %s connect %s (0), 0.5\n",soma_name,axon_name);
    }
    					 /* connect primary neurites */

    fprintf(fp,"\t\t\t\t\t/* connect primary neurites */\n");
    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

	fprintf(fp, "    %s connect %s[0] (0), 0.5\n",
		    soma_name, neurite_ptr->name);
    }
    fprintf(fp,"\n\n");
						 /* neurite geometry */

    fprintf(fp,"\t\t\t\t\t/* neurite geometry*/\n");
    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {


	fprintf(fp, for_loop_i, 0, neurite_ptr->num_sections-1);
	fprintf(fp, "        %s[i] {\n", neurite_ptr->name);
	fprintf(fp, "            nseg = fscan()\n");
	fprintf(fp, "            pt3dclear()\n");
	fprintf(fp, "            %s", for_loop_j );
	fprintf(fp, "                %s\n", read_3D_points);
	fprintf(fp, "            }\n");
	fprintf(fp, "        }\n");
	fprintf(fp, "    }\n\n");
    }
    fprintf(fp,"\n\n");

    					       /* branching topology */

    fprintf(fp,"\t\t\t\t\t/* branching topology*/\n");
    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

	fprintf(fp, for_loop_i, 1, neurite_ptr->num_sections-1);
	fprintf(fp, "        %s[fscan()] connect %s[i] (0), fscan()\n",
			     neurite_ptr->name, neurite_ptr->name);
	fprintf(fp, "    }\n\n");
    }

    fprintf(fp,"}\n\n");

    fprintf(fp,"geometry()\n");


    fprintf(fp,"\n\n");

    append_data (fp, gobi_fp, gobi_ln_fp, neuron_ptr, neurite_list);

					  /* NULL geometry procedure */
    fprintf(fp, "\n/*%s*/\n", sepline);
    fprintf(fp,"proc geometry() { \n\n");
    fprintf(fp,"\t/* NULL geometry procedure: keeps the user from\n");
    fprintf(fp,"\tcalling the geometry procedure in isolation. */\n\n");
    fprintf(fp,
    "    printf(\"\\nYou must re-read the entire geometry\\n\")\n");
    fprintf(fp,
    "    printf(\"file to execute geometry().\\n\\n\")\n");
    fprintf(fp,"}\n\n");
}


/*-------------------------------------------------------------------*/
static int write_multi_neuron_geometry( fp, gobi_fp, gobi_ln_fp,
			          neuron_ptr, 
			          neurite_list )


	/* Writes to fp a "geometry" subroutine in hoc syntax, using
	the info in the *neuron_ptr and *neurite_list.  Returns TRUE 
	if error; FALSE otherwise. */


    FILE      * fp,
              * gobi_fp,
              * gobi_ln_fp;
    NEURON    * neuron_ptr; 
    NEURITE   * neurite_list;


{
				       /*----- functions called -----*/
				       /*----- extern variables -----*/
				       /*----- local  variables -----*/
    static char for_loop_j[] = "for j = 1, fscan() {\n",
                read_3D_points[] =
		"pt3dadd(fscan(),fscan(),fscan(),fscan())";

    int			neurite_i,
                        section_i, 
                        xyzd_i,
                        i, procedure_counter, lines_so_far, blocks_so_far;

    NEURITE	      * neurite_ptr,
		      * neurite_list_end;
    SECTION	      * section_ptr,
		      * section_list_end;
    XYZD	      * xyzd_ptr,
		      * xyzd_list_end;
    char     soma_name[100], axon_name[100], dend_name[100], temp_buf[100];
    printf ("Inside write_multi_neuron_geometry.\n");

				       /*----- start function -------*/

    neurite_list_end = neurite_list + neuron_ptr->num_primary_neurites;

                                       /* handle cell names if present */

    if (neuron_ptr->cell_name != NULL) {
      sprintf(soma_name,"%s_%s",neuron_ptr->cell_name, SOMA_NAME);
      sprintf(axon_name,"%s_%s",neuron_ptr->cell_name, AXON_NAME);
      for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {
                 /*I had to create a temporary buffer for the new ->name, this 
                   was due to the fact that you can not sprintf to a string
                   that you need to use as an argument to sprintf.  so the
                   former line:
                    sprintf (neurite_ptr->name "%s_%s", neuron_ptr->cell_name, 
                             neurite_ptr->name) was crashing.         Ritter */
         sprintf(temp_buf, "%s_%s",neuron_ptr->cell_name, neurite_ptr->name);
         neurite_ptr->full_name = (char *) malloc (sizeof (temp_buf));
         strcpy (neurite_ptr->full_name, temp_buf);
       }
    } else {
      sprintf(soma_name,"%s",SOMA_NAME);
      sprintf(axon_name,"%s",AXON_NAME);
    }

    write_neuron_comment(fp, neuron_ptr);

    fprintf(fp,"/* If you want to use this file without the menu program\n");
    fprintf(fp,"   then you will need to remove the ropen() call in geometry(),\n");
    fprintf(fp,"   and assign \"%s\" equal to \"1\". *\\\n\n", neuron_ptr->cell_name);

    if (neuron_ptr->write_explicit_info) 
      write_neuron_explicit_info(fp, neuron_ptr, neurite_list);
    

    						  /* create sections */

    fprintf(fp,"\n\t\t\t\t\t/* create sections */\n");
    fprintf(fp,"create    %s[1][%s]",soma_name, neuron_ptr->cell_name);

                                                   /* create an axon */

    if (neuron_ptr->create_axon) {
      fprintf(fp,",\\\n          %s[1][%s]",axon_name, neuron_ptr->cell_name);
    }


    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

	fprintf(fp, ",\\\n          %s[%d][%s]",
		    neurite_ptr->full_name,
		    neurite_ptr->num_sections,
		    neuron_ptr->cell_name);
    }
    fprintf(fp,"\n\n");

    printf ("Writing main geometry procedure.\n");

					  /* main geometry procedure */

    fprintf(fp, "\n\n/*%s*/\n", sepline);
    fprintf(fp,"proc init_geometry () { local j\n\n");
    /* This routine would appear to be happy with 1 large forsec and multiple ifsecs,
       but this won't work because with a forsec you aren't assured the order in which
       you loop through all of the sections (same with forall) - in fact you will most
       probably loop through the sections in the reverse order that they were created.
       Thus the fscan() calls would get mixed up. */


    fprintf(fp, "\t/*---------- Coordinates and Diameters: soma -------------------*/\n\n");

    fprintf(fp, "\t%s[0][$1] {\n", soma_name);
    fprintf(fp, "\t  nseg = 1\n");
    fprintf(fp, "\t  pt3dclear()\n");
    fprintf(fp, "\t  %s", for_loop_j );
    fprintf(fp, "%\t\t%s\n", read_3D_points);
    fprintf(fp, "\t  }\n");
    fprintf (fp, "\t}\n\n");

                                            /* minimal axon geometry */
    
    if (neuron_ptr->create_axon) {
      fprintf(fp, "\t/*---------- Coordinates and Diameters: axon -------------------*/\n\n");
      fprintf(fp, "\t%s[0][$1] {\n",axon_name);
      fprintf(fp, "\t  nseg = 1\n");
      fprintf(fp, "\t  pt3dclear()\n");
      fprintf(fp, "\t  pt3dadd(0,0,0,1e-4)\n");
      fprintf(fp, "\t  pt3dadd(0,0,1e-4,1e-4)\n");
      fprintf(fp, "\t}\n\n");
    }
 
						 /* neurite geometry */
    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {


        fprintf(fp, "\t/*---------- Coordinates and Diameters: %s -----------*/\n\n", 
                 neurite_ptr->full_name);

        section_list_end = neurite_ptr->section_list +
			   neurite_ptr->num_sections;
	

	fprintf(fp, "\tfor i = 0, %d {\n", neurite_ptr->num_sections - 1);
	fprintf(fp, "\t  %s[i][$1] {\n", neurite_ptr->full_name);
	fprintf(fp, "\t    nseg = fscan()\n");
	fprintf(fp, "\t    pt3dclear()\n");
	fprintf(fp, "\t    %s", for_loop_j );
	fprintf(fp, "\t\t  %s\n", read_3D_points);
	fprintf(fp, "\t    }\n"); /* for i loop */
	fprintf(fp, "\t  }\n"); /* for j loop */
	fprintf(fp, "\t}\n\n");
      }

    fprintf(fp, "}\t/* Close the function init_geometry() */\n\n\n\n");
    
    fprintf(fp, "proc connect_geometry() { local j\n\n");

    if (neuron_ptr->create_axon) {
    fprintf(fp, "\t/*--------------------- Connecting axon --------------------*/\n\n");
      fprintf(fp, "\t%s[0][$1] connect %s[0][$1] (0), 0.5\n",soma_name,axon_name);
    }

    					 /* connect primary neurites */

    fprintf(fp, "\n\t/*--------------- Connecting primary neurites --------------*/\n\n");
    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

	fprintf(fp, "\t%s[0][$1] connect %s[0][$1] (0), 0.5\n",
		    soma_name, neurite_ptr->full_name);
    }

    					       /* branching topology */
    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {
    fprintf(fp, "\n\t/*--------------------- Connecting %s --------------------*/\n\n",
                                                        neurite_ptr->full_name);
    fprintf(fp, "\tfor j = 1, %d {\n", neurite_ptr->num_sections - 1);
    fprintf(fp, "\t\t%s[fscan()][$1] connect %s[j][$1] (0), fscan()\n",
	    neurite_ptr->full_name, neurite_ptr->full_name);
    fprintf(fp, "\t}\n");
  }
    fprintf(fp, "}\t/* Closes procedure connect_geometry() */\n\n\n\n");

                                                 /*geometry procedure*/
    fprintf(fp, "/*----------------------- geometry() ----------------------*/\n\n");
    fprintf(fp, "proc geometry() { local i\n\n");
    fprintf(fp, "\tfor i = 0, (%s - 1) {\n", neuron_ptr->cell_name);
    fprintf(fp, "\t\tropen(\"geom_data.tmp\")\n"); /* geom_data_file is set in neuron.hoc */
    fprintf(fp, "\t\tinit_geometry(i)\n\t\tconnect_geometry(i)\n");
    fprintf(fp, "\t\tropen()\t/* Close the geometry data file */\n");
    fprintf(fp, "\t}\n");
    fprintf(fp, "}\t/* Closes procedure geometry() */\n\n");


    fprintf(fp, "\ngeometry()\t/* Call the function geometry() */\n\n");

					  /* NULL geometry procedure */
    fprintf(fp, "\n/*%s*/\n", sepline);
    fprintf(fp,"proc geometry() { \n\n");
    fprintf(fp,"\t/* NULL geometry procedure: keeps the user from\n");
    fprintf(fp,"\tcalling the geometry procedure in isolation. */\n\n");
    fprintf(fp,
    "    printf(\"\\nYou must re-read the entire geometry\\n\")\n");
    fprintf(fp,
    "    printf(\"file to execute geometry().\\n\\n\")\n");
    fprintf(fp,"}\n\n");

    /* used as separator for splitting the file in neuron/SRC/menu.c */
    fprintf(fp, "MULTI NEURON DATA VALUES:\n");

append_data (fp, gobi_fp, gobi_ln_fp, neuron_ptr, neurite_list);

}

/*----------------------------------------------------------------------*/

static int append_data (fp, gobi_fp, gobi_ln_fp, neuron_ptr, neurite_list) 

     FILE    * fp,
             * gobi_fp,
             * gobi_ln_fp;
     NEURON  * neuron_ptr;
     NEURITE * neurite_list;

{
				       /*----- functions called -----*/
				       /*----- extern variables -----*/
				       /*----- local  variables -----*/

    int			neurite_i,
                        section_i, 
                        xyzd_i,
                        point_counter;

    NEURITE	      * neurite_ptr,
		      * neurite_list_end;
    SECTION	      * section_ptr,
		      * section_list_end;
    XYZD	      * xyzd_ptr,
		      * xyzd_list_end;


    				       /*----- start function -------*/
    printf ("Appending data.\n");

                                           /* Dump Data Points */
    point_counter = 1;

    neurite_list_end = neurite_list + neuron_ptr->num_primary_neurites;

    fprintf(fp, "SOMA COORDINATES AND DIAMETERS:\n");

    fprintf(fp, "\n    %d\n",
		neuron_ptr->num_soma_3D_points);

    xyzd_list_end = neuron_ptr->soma_3D_list +
		    neuron_ptr->num_soma_3D_points;

    for( xyzd_ptr = neuron_ptr->soma_3D_list;
	 xyzd_ptr < xyzd_list_end;
	 xyzd_ptr++ ) {

	fprintf(fp, 
	"%        8.5g %8.5g %8.5g %8.5g\n",
		xyzd_ptr->x,
		xyzd_ptr->y,
		xyzd_ptr->z,
		xyzd_ptr->diam);

	if (neuron_ptr->make_gobi_file){
	  fprintf(gobi_fp, "%8.5g\t%8.5g\t%8.5g\n", 
		xyzd_ptr->x,
		xyzd_ptr->y,
		xyzd_ptr->z);
	  point_counter++;
	}

      }
	
    fprintf(fp,"\n\n");

    fprintf(fp, "NEURITE COORDINATES AND DIAMETERS:\n");

/*    point_counter = 1;*/

    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

        section_list_end = neurite_ptr->section_list +
			   neurite_ptr->num_sections;

	for ( section_ptr = neurite_ptr->section_list;
	      section_ptr < section_list_end;
	      section_ptr++ ) {

	    fprintf(fp, "\n    %d %d\n",
			section_ptr->num_segments,
			section_ptr->num_xyzd_points);

	    xyzd_list_end = section_ptr->xyzd_list +
			    section_ptr->num_xyzd_points;

	    for( xyzd_ptr = section_ptr->xyzd_list;
		 xyzd_ptr < xyzd_list_end;
		 xyzd_ptr++) {

		fprintf(fp, 
		"%        8.5g %8.5g %8.5g %8.5g\n",
			xyzd_ptr->x,
			xyzd_ptr->y,
			xyzd_ptr->z,
			xyzd_ptr->diam);

		  if (neuron_ptr->make_gobi_file) {
		    fprintf(gobi_fp, "%8.5g\t%8.5g\t%8.5g\n", 
			  xyzd_ptr->x,
			  xyzd_ptr->y,
			  xyzd_ptr->z);
		    ++xyzd_ptr;
		    if (! (xyzd_ptr == xyzd_list_end)) {
		      fprintf(gobi_ln_fp, "%d %d\n", point_counter, 
			                      point_counter+1);
		    }
		    --xyzd_ptr;
		    point_counter++;
		  }
	      }
	  }
      }

    fprintf(fp,"\n\n");

    fprintf(fp, "CONNECTIONS:\n\n");

    for ( neurite_ptr = neurite_list;
          neurite_ptr < neurite_list_end;
          neurite_ptr++ ) {

        section_list_end = neurite_ptr->section_list +
			   neurite_ptr->num_sections;

	for ( section_ptr = neurite_ptr->section_list+1;
	      section_ptr < section_list_end;
	      section_ptr++ ) {

	    fprintf(fp,"    %5d    %g\n",
		       section_ptr->parent_index,
		       section_ptr->origin_in_parent);

	}

        fprintf(fp,"\n");
    }

    return(ferror(fp));

  }


/*-------------------------------------------------------------------*/
static int write_cable_geometry( fp, 
			         neuron_ptr, 
			         cable_ptr )


	/* Writes to fp a "geometry" subroutine in hoc syntax, using
	the info in the *neuron_ptr and *cable_ptr.  Returns TRUE if 
	error; FALSE otherwise. */


    FILE      * fp; 
    NEURON    * neuron_ptr; 
    CABLE     * cable_ptr;


{
				       /*----- functions called -----*/
				       /*----- extern variables -----*/
				       /*----- local  variables -----*/

    static char for_loop[] = "    for (i=%d; i<=%d; i=i+1)\t{";
    char      * type_index;
    int 	first,
		last,
		i,
		num_lines_in_procedure,
		num_fcbs_procedures,
		num_dx_procedures,
		num_diam_procedures;
    double	first_diam;
    double	last_diam;
    double	slope;

    FCBS_ARGS         * fcbs_args_ptr,
		      * fcbs_list_end;
    DX_INTERVAL       * dx_interval_ptr,
   		      * dx_list_end;
    DIAM_INTERVAL     * diam_interval_ptr,
   		      * diam_list_end;

				       /*----- start function -------*/

					 /* check translation syntax */
    switch (neuron_ptr->destination) {
	case O_MULTI:
	    type_index = "[type]";
	    break;
	case O_CABLE:
	default:
	    type_index = "";
    }


    write_neuron_comment(fp, neuron_ptr);

    					       /* branching topology */
    num_fcbs_procedures = 0;
    fcbs_args_ptr = cable_ptr->fcbs_args_list;
    fcbs_list_end = fcbs_args_ptr + cable_ptr->num_fcbs_args;

    do {
        fprintf(fp, "\n\n/*%s*/\n", sepline);
        fprintf(fp,"proc init_topology_%d() {\n\n",
			num_fcbs_procedures);

	if (!num_fcbs_procedures)
            fprintf(fp,"    fcbs(0)\n");

        for (num_lines_in_procedure = 0;
		
             fcbs_args_ptr < fcbs_list_end
	     &&
	     num_lines_in_procedure < MAX_LINES_PER_PROC - 3;

	     num_lines_in_procedure++,
             fcbs_args_ptr++)

    	    fprintf(fp, "    fcbs(%d, %d, %d)\n", 
    	    fcbs_args_ptr->first_index,
    	    fcbs_args_ptr->last_index,
    	    fcbs_args_ptr->root_index);

        fprintf(fp,"}\n");
	num_fcbs_procedures++;
    } while (fcbs_args_ptr < fcbs_list_end);
	
    						  /* segment lengths */
    num_dx_procedures = 0;
    dx_interval_ptr = cable_ptr->dx_interval_list;
    dx_list_end = dx_interval_ptr + cable_ptr->num_dx_intervals;

    do {
        fprintf(fp, "\n\n/*%s*/\n", sepline);
        fprintf(fp,"proc init_dx_%d() { local i\n\n",
			num_dx_procedures);

        for (num_lines_in_procedure = 0;

             dx_interval_ptr < dx_list_end
	     &&
	     num_lines_in_procedure < MAX_LINES_PER_PROC - 2;

	     num_lines_in_procedure++,
             dx_interval_ptr++) {

	    if ( (first = dx_interval_ptr->first_index)
		    ==
    	         (last = dx_interval_ptr->last_index) ) {
    	        fprintf(fp,"    dx%s[%d] = %g\n", type_index,
		first, dx_interval_ptr->dx);
	    }
	    else if ( first + 1 == last  ) {
    	        fprintf(fp,"    dx%s[%d] = %g\n", type_index,
		first, dx_interval_ptr->dx);
    	        fprintf(fp,"    dx%s[%d] = %g\n", type_index,
		last, dx_interval_ptr->dx);
	        num_lines_in_procedure++;
	    }
	    else {
	        fprintf(fp, for_loop, first, last);
    	        fprintf(fp," dx%s[i] = %g }\n", type_index,
    	        dx_interval_ptr->dx);
	    }

        }

        fprintf(fp,"}\n");
	num_dx_procedures++;
    } while (dx_interval_ptr < dx_list_end);

    						/* segment diameters */

    num_diam_procedures = 0;
    diam_interval_ptr = cable_ptr->diam_interval_list;
    diam_list_end = diam_interval_ptr + cable_ptr->num_diam_intervals;

    do {
        fprintf(fp, "\n\n/*%s*/\n", sepline);
        fprintf(fp,"proc init_diam_%d() { local i\n\n",
			num_diam_procedures);

        for (num_lines_in_procedure = 0;

             diam_interval_ptr < diam_list_end
	     &&
	     num_lines_in_procedure < MAX_LINES_PER_PROC - 2;

	     num_lines_in_procedure++,
             diam_interval_ptr++) {

	    first      = diam_interval_ptr->first_index;
	    last       = diam_interval_ptr->last_index;
	    first_diam = diam_interval_ptr->first_diam;
	    last_diam  = diam_interval_ptr->last_diam;

	    if ( first == last ) {
    	        fprintf(fp,"    diam%s[%d] = %g\n", type_index,
		first, (first_diam + last_diam)/2 );
	    }
	    else if ( first + 1 == last  ) {
    	        fprintf(fp,"    diam%s[%d] = %g\n", type_index,
		first, first_diam);
    	        fprintf(fp,"    diam%s[%d] = %g\n", type_index,
		last, last_diam);
	        num_lines_in_procedure++;
	    }
	    else if (first_diam == last_diam) {
	        fprintf(fp, for_loop, first, last);
    	        fprintf(fp," diam%s[i] = %g }\n", type_index,
		first_diam);
	    }
	    else {
	        slope = (last_diam - first_diam)/(last-first);
	        fprintf(fp, for_loop, first, last);
    	        fprintf(fp," diam%s[i] = %g %c %g*(i-%d) }\n",
		type_index,
	        first_diam, slope<0 ? '-' : '+', fabs(slope), first);
	    }

        }

        fprintf(fp,"}\n");
	num_diam_procedures++;
    } while (diam_interval_ptr < diam_list_end);

					  /* main geometry procedure */
    fprintf(fp, "\n\n/*%s*/\n", sepline);
    fprintf(fp,"proc geometry() {\n\n");

    for (i = 0; i<num_fcbs_procedures; i++)
    	fprintf(fp, "    init_topology_%d()\n\n", i);

    for (i = 0; i<num_dx_procedures; i++)
    	fprintf(fp, "    init_dx_%d()\n\n", i);

    for (i = 0; i<num_diam_procedures; i++)
    	fprintf(fp, "    init_diam_%d()\n\n", i);

						    /* soma geometry */

    fprintf(fp,"\t\t\t\t\t/* cylindrical soma geometry */\n");
    fprintf(fp,"    diam%s[0] = %g\n", type_index,
    		neuron_ptr->soma_diam);
    fprintf(fp,"    dx%s[0]   = %g\n", type_index, 
		neuron_ptr->soma_length);
    fprintf(fp,"    ra%s      = %g\n", type_index,
		neuron_ptr->cytoplasmic_resistivity);
    fprintf(fp,"    farea(3)\n");
    fprintf(fp,"    area%s[0] = PI*diam%s[0]*dx%s[0]\n", 
		type_index, type_index, type_index );
    fprintf(fp,"    fcbs()\n\n");
    fprintf(fp,"}\n\n");

    return(ferror(fp));

}


/*-------------------------------------------------------------------*/
int write_neuron_description(fp, neuron_ptr)

	/* Writes to fp a text description of the neuron and its
	translation. Returns TRUE if error, FALSE otherwise. */

    FILE      * fp;
    NEURON    * neuron_ptr;

{
				       /*----- functions called -----*/

				       /*----- extern variables -----*/
    extern char version[];

				       /*----- local  variables -----*/

				       /*----- start function -------*/

    fprintf(fp,
    "%s %s translated %24.24s by ntscable %s\n",
    neuron_ptr->input_filename,
    neuron_ptr->input_version,
    neuron_ptr->time_of_translation,
    version);
    fprintf(fp, "source file syntax: %s\n", 
	source_syntax[(neuron_ptr->source)]);
    fprintf(fp, "output file syntax: %s\n", 
	output_syntax[(neuron_ptr->destination)]);

    fprintf(fp,
    "%s: diameter = %g um  length = %g um  area = %g um2\n",
    SOMA_NAME,neuron_ptr->soma_diam, neuron_ptr->soma_length,
    neuron_ptr->soma_area);

    fprintf(fp, "      %d three-D points", 
    	neuron_ptr->num_soma_3D_points);
    if (neuron_ptr->num_soma_outline_points) {
        fprintf(fp, "; %d outline points",
    	neuron_ptr->num_soma_outline_points);
    }
    if (neuron_ptr->first_soma_point
	||
        neuron_ptr->last_soma_point) {
	fprintf(fp,
	" numbered %d-%d",
	neuron_ptr->first_soma_point,
	neuron_ptr->last_soma_point);
    }
    fprintf(fp, "\n");

    if (neuron_ptr->num_soma_outline_points) {
        fprintf(fp, "      outline diameter = %g um%s\n",
    	neuron_ptr->soma_outline_diam,
        (neuron_ptr->average_outline_z_coords) ?
	"; z coordinates averaged" : "" );
    }

    fprintf(fp,
    "%d primary neurites\n",
    neuron_ptr->num_primary_neurites);

    fprintf(fp,
    "%d branches totaling %g um in length, %g um2 in area\n",
    neuron_ptr->num_branches,
    neuron_ptr->total_length_of_neurites,
    neuron_ptr->total_area_of_neurites);

    fprintf(fp,
    "%d tree points translated to %d segments (%d requested)\n",
    neuron_ptr->num_tree_points,
    neuron_ptr->num_segments_used,
    neuron_ptr->num_segments_requested);

    fprintf(fp,
    "Neurites divided into segments of equal dx ");

    switch (neuron_ptr->accuracy_level) {
	case BETWEEN_POINTS:
            fprintf(fp,
	    "between adjacent digitized\npoints.\n");
	    break;
	case BETWEEN_BRANCHPOINTS:
            fprintf(fp,
	    "between adjacent digitized\nbranch points.\n");
	    break;
	case WHOLE_BRANCH:
            fprintf(fp,
	    "over each entire branch.\n");
    }

    fprintf(fp,
    "Segment length constrained to be < %g um",
    neuron_ptr->max_dx);
    if (neuron_ptr->max_dx_over_mean_diam > 0.0)  {
        fprintf(fp,
	", or < %g times\n",
        neuron_ptr->max_dx_over_mean_diam );
        fprintf(fp,
	"mean diameter of translation interval, whichever is less.\n");
    }
    else {
        fprintf(fp,".\n");
    }

    fprintf(fp, "%s", neuron_ptr->comment);

    return(ferror(fp));

}




/*-------------------------------------------------------------------*/
static int write_neuron_comment(fp, neuron_ptr)

	/* Writes to fp a HOC-style comment describing the neuron and 
	its translation. Returns TRUE if error, FALSE otherwise. */

    FILE      * fp;
    NEURON    * neuron_ptr;


{
				       /*----- functions called -----*/
    char * strchr();

				       /*----- extern variables -----*/

				       /*----- local  variables -----*/
    char  * ptr;

				       /*----- start function -------*/

    			     /* delete any C-style comment terminators 
    			     from the comment */

    ptr = neuron_ptr->comment;

    while( *ptr && (ptr = strchr(ptr, '*')) )
	if (*(ptr+1) == '/') 
	    strcpy(ptr, ptr+2);
	else
	    ptr++;

    		       /* Write comment describing the source file and
    		       the results of translation.  Include leading 
    		       comment from the source file. */

    fprintf(fp, "/*%s\n", sepline);
    fprintf(fp, "%cHeader%c\n", '$', '$');

    write_neuron_description(fp, neuron_ptr);

    fprintf(fp, "%s*/\n", sepline);
    return(ferror(fp));

}

/*-- ZFM modification ------------------------------------------------------*/
static int write_neuron_explicit_info(fp, neuron_ptr, neurite_list)

	/* Writes to fp a set of hoc-syntax definitions making 
	   information about the neuron explicitly accessible to 
	   the hoc user.  Returns TRUE if error, FALSE otherwise. */

    FILE      * fp;
    NEURON    * neuron_ptr;
    NEURITE   * neurite_list;

{
				       /*----- functions called -----*/
				       /*----- extern variables -----*/
				       /*----- local  variables -----*/
				       /*----- start function -------*/

  NEURITE * neurite_ptr,
          * neurite_list_end;
  int neurite_i, num_sections;


  neurite_list_end = neurite_list + neuron_ptr->num_primary_neurites;

  fprintf(fp,"\n\t\t\t\t\t/* make geometry explicit */\n");
  fprintf(fp, "primary_branches_cell = %d\n", neuron_ptr->num_primary_neurites);
  fprintf(fp, "branches_cell = %d\n", neuron_ptr->num_branches);
  fprintf(fp, "max_dx_cell = %g\n", neuron_ptr->max_dx);
  fprintf(fp, "points_cell = %d\n", neuron_ptr->num_tree_points);
  fprintf(fp, "nseg_cell = %d\n", neuron_ptr->num_segments_used);


  fprintf(fp, "double sections_cell[%d]\n", neuron_ptr->num_primary_neurites);
  neurite_i = 0;
  num_sections = 0;
  for ( neurite_ptr = neurite_list;
       neurite_ptr < neurite_list_end;
       neurite_ptr++ ) {
    fprintf(fp, "sections_cell[%d] = %d\n",neurite_i,neurite_ptr->num_sections);
    num_sections += neurite_ptr->num_sections;
    neurite_i++;
  }
        /* add for soma and axon (if present) */
  if (neuron_ptr->create_axon) {
    num_sections = num_sections + 2; 
  } else {
    num_sections++;
  }
  fprintf(fp, "nsec_cell = %d\n\n", num_sections);

  fprintf(fp, "strdef name_soma, name_section");
  if (neuron_ptr->create_axon) {
    fprintf(fp, ", name_axon\n");
  } else {
    fprintf(fp, "\n");
  }
  /*fprintf(fp, "name_soma = \"%s\"\nname_section = \"%s\"\n",(neuron_ptr->cell_name == NULL) ? SOMA_NAME : strncat(neuron_ptr->cell_name, "_", SOMA_NAME,SECTION_NAME);*/
  if (neuron_ptr->cell_name == NULL) {
    fprintf(fp, "name_soma = \"%s\"\nname_section = \"%s\"\n", SOMA_NAME, SECTION_NAME);
    if (neuron_ptr->create_axon) {
      fprintf(fp, "name_axon = \"%s\"\n", AXON_NAME);
      fprintf(fp, "use_axon = 1");
    } else {
    fprintf(fp, "use_axon = 0");
  }
  }

  else {
    fprintf(fp, "name_soma = \"%s_%s\"\nname_section = \"%s_%s\"\n", 
            neuron_ptr->cell_name, SOMA_NAME, neuron_ptr->cell_name, SECTION_NAME);
    if (neuron_ptr->create_axon) {
      fprintf(fp, "name_axon = \"%s_%s\"\n",neuron_ptr->cell_name, AXON_NAME);
      fprintf(fp, "use_axon = 1");
    } else {
    fprintf(fp, "use_axon = 0");
    }
  }
  return(ferror(fp));
}



