/*---------------------------------------------------------------------
$Header: /usr/local/src/tools/RCS/directio.c,v 1.2 1992/07/17 19:29:49 fisher Exp $

Routines for direct (unbuffered) terminal I/O.  The unbuffered mode
will be cbreak mode on a BSD system; it will be raw i/o with System V.

AD: modified for LINUX (SUSE 7.1 distribution)
---------------------------------------------------------------------*/

#include <stdio.h>
#include <ctype.h>
#include "jw.h"
#include "directio.h"

#if OPSYS == SYS_V
#include <termio.h>
#include <stropts.h>
#endif

#if OPSYS == BSD
#include <sgtty.h>
#endif

/*---------------------------------- PRIVATE DECLARATIONS ------------*/

#if OPSYS == SYS_V
#define STRUCT_TERMIO	struct termio
#define	LINE_DISCIPLINE	c_lflag
#define	GET_TERM_PARAMS	TCGETA
#define	SET_TERM_PARAMS	TCSETA
#define ECHO_MASK	ECHO | ECHOE | ECHOK | ECHONL
#endif

#if OPSYS == BSD
#define STRUCT_TERMIO	struct sgttyb
#define	LINE_DISCIPLINE	sg_flags
#define	GET_TERM_PARAMS TIOCGETP
#define	SET_TERM_PARAMS TIOCSETP
#define CBREAK_MASK	2		  /* sgttyb masks; see */
#define ECHO_MASK	8		  /* tty(4)	       */
#endif

PRIVATE STRUCT_TERMIO
		tty_cooked_params,
		tty_cbreak_params;


PRIVATE int	kbd_fdesc,
		scrn_fdesc,
		initialized = FALSE;

/*====================== SYSTEM-DEPENDENT CODE ======================*/

/*-------------------------------------------------------------------*/
void init_io_params()

	/* Gets default tty parameters and saves them in global struct
	variable tty_cooked_params; copies these to tty_cbreak_params
	and modifies the latter for c_break mode with no echo.  Also 
	initializes global file descriptors kbd_fdesc and scrn_fdesc.
	These globals are used in I/O routines io_cbreak(), 
	io_cooked(), keypressed(), getkey(), and echokey().  See 
	tty(4) for more info. */

{
				       /*----- extern variables -----*/
    extern STRUCT_TERMIO
		tty_cooked_params,
		tty_cbreak_params;

    extern int	kbd_fdesc,
		scrn_fdesc;
				       /*----- start function -------*/


					/* Get current tty parameters;
					0 is standard input.   */

    ioctl(0, GET_TERM_PARAMS, &tty_cooked_params);
	     						     /* copy */
    tty_cbreak_params = tty_cooked_params;

				      /* set up for unbuffered input */
#if OPSYS == SYS_V
    tty_cbreak_params.LINE_DISCIPLINE &= ~ICANON;
    tty_cbreak_params.c_cc[VMIN] = 1;
    tty_cbreak_params.c_cc[VTIME] = 0;
#endif

#if OPSYS == BSD
    tty_cbreak_params.LINE_DISCIPLINE |= CBREAK_MASK;
#endif

	       					       /* clear echo */
    tty_cbreak_params.LINE_DISCIPLINE &= ~ECHO_MASK;

    scrn_fdesc = 1;			   /* screen file descriptor */
    kbd_fdesc  = 0;			 /* keyboard file descriptor */
    initialized = TRUE;

}

/*-------------------------------------------------------------------*/
void purge_keyboard()

	/* Reads and discards all characters waiting to be read from
	keyboard. */

{
				       /*----- functions called -----*/
    char 	getkey();
				       /*----- extern variables -----*/
    extern int	kbd_fdesc;
				       /*----- local  variables -----*/
    long	num_chars_ready;
				       /*----- start function -------*/

#if OPSYS == BSD
    ioctl(kbd_fdesc, FIONREAD, &num_chars_ready);
    while(num_chars_ready--)
	getkey();
#endif

#if OPSYS == SYS_V
    ioctl(kbd_fdesc, TCFLSH, 0);
#endif

}

/*-------------------------------------------------------------------*/
int keypressed()

	/* Returns TRUE if there is keyboard input waiting to be read;
	otherwise returns FALSE.  Works both in cbreak and cooked
	modes. */

{
				       /*----- extern variables -----*/
    extern int	  kbd_fdesc;
    extern STRUCT_TERMIO
		  tty_cbreak_params;
				       /*----- local  variables -----*/
    long	  num_chars_ready;

#if OPSYS == SYS_V
    STRUCT_TERMIO params_on_entry;
    char 	  c;
#endif
				       /*----- start function -------*/

#if OPSYS == BSD
    ioctl(kbd_fdesc, FIONREAD, &num_chars_ready);
#endif

#if OPSYS == SYS_V

    ioctl(0, GET_TERM_PARAMS, &params_on_entry);
    io_cbreak();
    ioctl(0, I_NREAD, &num_chars_ready);
    ioctl(0, SET_TERM_PARAMS, &params_on_entry);

#endif


    return(num_chars_ready > 0);
}



/*========================== PORTABLE CODE ==========================*/

/*-------------------------------------------------------------------*/
char getkey()

	/* Waits for, reads and returns char from keyboard.  Reads
	char without line-editing and without echo to screen if in
	cbreak mode. */

{
				       /*----- extern variables -----*/
    extern int	kbd_fdesc;

				       /*----- local  variables -----*/
    char	c;
				       /*----- start function -------*/
    read(kbd_fdesc, &c, 1);
    return(c);
}

/*-------------------------------------------------------------------*/
int return_pressed()

	/* Returns TRUE if user presses RETURN key; otherwise returns
	FALSE.  Must be in unbuffered mode. */

{
				       /*----- functions called -----*/
    int		keypressed();
    char	getkey();
				       /*----- start function -------*/
    if (keypressed())
	return ( getkey() == '\n' );
    else
	return (FALSE);
}

/*-------------------------------------------------------------------*/
void wait_for_keypress()

	/* Prompts user to press a key, then waits until key is 
	pressed.  Uses cbreak mode while waiting, but then restores to 
	mode that was in effect on entry. */

{
				       /*----- functions called -----*/
    char 	getkey();
				       /*----- extern variables -----*/
    extern int	kbd_fdesc,
		scrn_fdesc;
				       /*----- local  variables -----*/

    STRUCT_TERMIO params_on_entry;
    char c;
				       /*----- start function -------*/

						/* Get current tty
						parameters; 0 is 
						standard input.   */
    ioctl(0, GET_TERM_PARAMS, &params_on_entry);

    io_cbreak();				       /* set cbreak */

    printf("\n(press a key to continue) ");
    fflush(stdout);
    c = getkey();
    printf("\r                         \r");
    fflush(stdout);

    ioctl(kbd_fdesc, SET_TERM_PARAMS, &params_on_entry);

}

/*-------------------------------------------------------------------*/
int user_presses(target_key)

	/* Waits until key is pressed.  Uses cbreak mode while 
	waiting, but then restores to mode that was in effect on entry.
	Returns TRUE if user presses the key specified by target_key; 
	FALSE otherwise. */

    char	target_key;

{
				       /*----- functions called -----*/
    char 	getkey();
				       /*----- extern variables -----*/
    extern int	kbd_fdesc,
		scrn_fdesc;
				       /*----- local  variables -----*/

    STRUCT_TERMIO params_on_entry;
    char c;
				       /*----- start function -------*/

						/* Get current tty
						parameters; 0 is 
						standard input.   */
    ioctl(0, GET_TERM_PARAMS, &params_on_entry);	

    io_cbreak();				       /* set cbreak */
    c = getkey();
    ioctl(kbd_fdesc, SET_TERM_PARAMS, &params_on_entry);
    return(c==target_key);

}

/*-------------------------------------------------------------------*/
int user_quits_by_pressing(quit_key)

	/* Prompts user to press quit_key to quit, or any other key to
	continue. Waits until key is pressed, then returns TRUE if 
	quit_key or its uppercase equivalent was pressed; returns
	FALSE otherwise.  Uses cbreak mode while waiting, but then 
	restores to mode that was in effect on entry. */

    char	quit_key;
{
				       /*----- functions called -----*/
    char 	getkey();
				       /*----- extern variables -----*/
    extern int	kbd_fdesc,
		scrn_fdesc;
				       /*----- local  variables -----*/

    STRUCT_TERMIO params_on_entry;
    char c;
				       /*----- start function -------*/

						/* Get current tty
						parameters; 0 is 
						standard input.   */
    ioctl(0, GET_TERM_PARAMS, &params_on_entry);	

    io_cbreak();				       /* set cbreak */

    if (quit_key < ' ')
	quit_key = '\n';

    printf("\n(press ");
    if (quit_key=='\n')
	printf("RETURN");
    else
	printf("%c", quit_key);

    printf(" to quit, or any other key to continue) ");
    fflush(stdout);
    c = getkey();
    printf("\r                                                    \r");
    fflush(stdout);

    ioctl(kbd_fdesc, SET_TERM_PARAMS, &params_on_entry);

    if (islower(quit_key))
	quit_key=toupper(quit_key);

    if (islower(c))
	c = toupper(c);

    return(c == quit_key);
}


/*-------------------------------------------------------------------*/
void io_cbreak()

	/* Changes terminal I/O parameters so that chars can be read
	directly from keyboard without echo to screen.  See tty(4). */

{
				       /*----- extern variables -----*/
    extern STRUCT_TERMIO
		tty_cbreak_params;

    extern int	kbd_fdesc;
				       /*----- start function -------*/
    if (! initialized)
	init_io_params();

    ioctl(kbd_fdesc, SET_TERM_PARAMS, &tty_cbreak_params);
}

/*-------------------------------------------------------------------*/
void io_cooked()

	/* Changes terminal I/O parameters so that chars are read
	only after normal buffered line editing.  See tty(4). */

{
				       /*----- extern variables -----*/
    extern STRUCT_TERMIO
		tty_cooked_params;

    extern int	kbd_fdesc;
				       /*----- start function -------*/
    if (! initialized)
	init_io_params();

    ioctl(kbd_fdesc, SET_TERM_PARAMS, &tty_cooked_params);
}

/*-------------------------------------------------------------------*/
void echokey(c)

	/* Writes char c to screen.  Works in both cooked and cbreak 
	modes. */

    char c;
{
				       /*----- extern variables -----*/
    extern int	scrn_fdesc;
				       /*----- start function -------*/
    write(scrn_fdesc, &c, 1);
}
